from __future__ import annotations

import asyncio
from datetime import timedelta
from typing import TYPE_CHECKING, Any, TypedDict

from temporalio import workflow as temporal_workflow
from temporalio.common import RetryPolicy
from typing_extensions import NotRequired, Unpack

if TYPE_CHECKING:
    from collections.abc import Callable

    from temporalio.types import CallableAsyncType
from temporalio.workflow import ChildWorkflowCancellationType, ParentClosePolicy

from .workflow import WorkflowLogger

log = WorkflowLogger()

get_external_agent_handle = temporal_workflow.get_external_workflow_handle
agent_info = temporal_workflow.info
continue_as_new = temporal_workflow.continue_as_new
condition = temporal_workflow.wait_condition
import_functions = temporal_workflow.unsafe.imports_passed_through
uuid = temporal_workflow.uuid4

__all__ = [
    "RetryPolicy",
    "agent_info",
    "condition",
    "continue_as_new",
    "get_external_agent_handle",
    "import_functions",
    "log",
    "uuid",
]


class StepKwargs(TypedDict):
    function: Any
    function_input: NotRequired[Any]
    task_queue: NotRequired[str]
    retry_policy: NotRequired[RetryPolicy]
    schedule_to_close_timeout: NotRequired[timedelta]


class ChildKwargs(TypedDict, total=False):
    workflow: Any
    workflow_id: str
    workflow_input: Any
    agent: Any
    agent_id: str
    agent_input: Any
    task_queue: str
    cancellation_type: ChildWorkflowCancellationType
    parent_close_policy: ParentClosePolicy
    execution_timeout: timedelta


class Agent:
    def defn(self, *args: Any, **kwargs: Any) -> Any:
        return temporal_workflow.defn(*args, **kwargs, sandboxed=False)

    def state(
        self, fn: Any, name: str | None = None, description: str | None = None
    ) -> Any:
        return temporal_workflow.query(fn, name=name, description=description)

    def event(
        self, fn: Any, name: str | None = None, description: str | None = None
    ) -> Any:
        return temporal_workflow.update(fn, name=name, description=description)

    def run(self, fn: CallableAsyncType) -> CallableAsyncType:
        return temporal_workflow.run(fn)

    def condition(
        self, fn: Callable[[], bool], timeout: timedelta | None = None
    ) -> None:
        return temporal_workflow.wait_condition(fn, timeout=timeout)

    async def step(
        self,
        **kwargs: Unpack[StepKwargs],
    ) -> Any:
        function = kwargs.get("function")
        function_input = kwargs.get("function_input")
        task_queue = kwargs.get("task_queue", "restack")
        retry_policy = kwargs.get("retry_policy")
        schedule_to_close_timeout = kwargs.get(
            "schedule_to_close_timeout", timedelta(minutes=2)
        )
        engine_id = self.get_engine_id_from_client()
        return await temporal_workflow.execute_activity(
            activity=function,
            args=(function_input,) if function_input is not None else (),
            task_queue=f"{engine_id}-{task_queue}",
            schedule_to_close_timeout=schedule_to_close_timeout,
            retry_policy=retry_policy,
        )

    async def child_start(
        self,
        **kwargs: Unpack[ChildKwargs],
    ) -> Any:
        task_queue = kwargs.get("task_queue", "restack")
        cancellation_type = kwargs.get(
            "cancellation_type",
            ChildWorkflowCancellationType.WAIT_CANCELLATION_COMPLETED,
        )
        parent_close_policy = kwargs.get(
            "parent_close_policy",
            ParentClosePolicy.TERMINATE,
        )
        execution_timeout = kwargs.get("execution_timeout")
        workflow = kwargs.get("workflow")
        workflow_id = kwargs.get("workflow_id")
        agent = kwargs.get("agent")
        agent_id = kwargs.get("agent_id")
        workflow_input = kwargs.get("workflow_input")
        agent_input = kwargs.get("agent_input")

        if not workflow and not agent:
            error_message = "Either workflow or agent must be provided."
            log.error(error_message)
            raise ValueError(error_message)

        if workflow and agent:
            error_message = "Either workflow or agent must be provided, but not both."
            log.error(error_message)
            raise ValueError(error_message)

        engine_id = self.get_engine_id_from_client()

        return await temporal_workflow.start_child_workflow(
            workflow=workflow.run or agent.run,
            args=[workflow_input or agent_input]
            if workflow_input or agent_input
            else [],
            id=self.add_engine_id_prefix(engine_id, workflow_id or agent_id),
            task_queue=f"{engine_id}-{task_queue}",
            memo={"engineId": engine_id},
            search_attributes={"engineId": [engine_id]},
            cancellation_type=cancellation_type,
            parent_close_policy=parent_close_policy,
            execution_timeout=execution_timeout,
        )

    async def child_execute(
        self,
        **kwargs: Unpack[ChildKwargs],
    ) -> Any:
        workflow = kwargs.get("workflow")
        workflow_id = kwargs.get("workflow_id")
        agent = kwargs.get("agent")
        agent_id = kwargs.get("agent_id")
        workflow_input = kwargs.get("workflow_input")
        agent_input = kwargs.get("agent_input")
        task_queue = kwargs.get("task_queue", "restack")
        cancellation_type = kwargs.get(
            "cancellation_type",
            ChildWorkflowCancellationType.WAIT_CANCELLATION_COMPLETED,
        )
        parent_close_policy = kwargs.get(
            "parent_close_policy",
            ParentClosePolicy.TERMINATE,
        )
        execution_timeout = kwargs.get("execution_timeout")
        if not workflow and not agent:
            error_message = "Either workflow or agent must be provided."
            log.error(error_message)
            raise ValueError(error_message)

        if workflow and agent:
            error_message = "Either workflow or agent must be provided, but not both."
            log.error(error_message)
            raise ValueError(error_message)

        engine_id = self.get_engine_id_from_client()

        return await temporal_workflow.execute_child_workflow(
            workflow=workflow.run or agent.run,
            args=[workflow_input or agent_input]
            if workflow_input or agent_input
            else [],
            id=self.add_engine_id_prefix(engine_id, workflow_id or agent_id),
            task_queue=f"{engine_id}-{task_queue}",
            memo={"engineId": engine_id},
            search_attributes={"engineId": [engine_id]},
            cancellation_type=cancellation_type,
            parent_close_policy=parent_close_policy,
            execution_timeout=execution_timeout,
        )

    async def sleep(self, seconds: int) -> Any:
        return await asyncio.sleep(seconds)

    def get_engine_id_from_client(self) -> Any:
        return temporal_workflow.memo_value("engineId", "local")

    def add_engine_id_prefix(self, engine_id: str, agent_id: str) -> str:
        if agent_id.startswith(f"{engine_id}-"):
            return agent_id
        return f"{engine_id}-{agent_id}"


agent = Agent()



import json
from typing import Any

import websockets
from temporalio import activity
from temporalio.exceptions import ApplicationError

from .observability import log_with_context, logger

activity.logger.logger = logger


class ActivityLogger:
    """Wrapper for activity logger that ensures proper context and formatting."""

    def __init__(self) -> None:
        self._logger = activity.logger

    def _log(self, level: str, message: str, **kwargs: Any) -> None:
        try:
            activity.info()
            getattr(self._logger, level)(
                message,
                extra={"extra_fields": {**kwargs, "client_log": True}},
            )
        except RuntimeError:
            log_with_context(level.upper(), message, **kwargs)

    def debug(self, message: str, **kwargs: Any) -> None:
        self._log("debug", message, **kwargs)

    def info(self, message: str, **kwargs: Any) -> None:
        self._log("info", message, **kwargs)

    def warning(self, message: str, **kwargs: Any) -> None:
        self._log("warning", message, **kwargs)

    def error(self, message: str, **kwargs: Any) -> None:
        self._log("error", message, **kwargs)

    def critical(self, message: str, **kwargs: Any) -> None:
        self._log("critical", message, **kwargs)

    def exception(self, message: str, **kwargs: Any) -> None:
        """Log an exception with traceback (equivalent to `logging.exception`)."""
        kwargs["exc_info"] = True
        self._log("exception", message, **kwargs)


log = ActivityLogger()

FunctionFailure = ApplicationError
function_info = activity.info
heartbeat = activity.heartbeat
function = activity

__all__ = [
    "FunctionFailure",
    "function_info",
    "heartbeat",
    "log",
]


def current_workflow() -> Any:
    return activity.Context.current().info


async def stream_to_websocket(api_address: str, data: Any) -> Any:
    """Stream data to Restack Engine WebSocket API endpoint.

    Args:
        api_address (str): The address of the Restack Engine API.
        data (Any): The streamed data from an OpenAI-compatible API or a JSON dict.

    Returns:
        str: The final combined response as a string.

    """
    info = {
        "activityId": activity.info().activity_id,
        "workflowId": activity.info().workflow_id,
        "runId": activity.info().workflow_run_id,
        "activityType": activity.info().activity_type,
        "taskQueue": activity.info().task_queue,
    }

    protocol = "ws" if api_address.startswith("localhost") else "wss"
    websocket_url = f"{protocol}://{api_address}/stream/ws/workflow/history?workflowId={info['workflowId']}&runId={info['runId']}"

    try:
        async with websockets.connect(websocket_url) as websocket:
            try:
                # Send initial message
                await websocket.send(str(info))
                heartbeat(info)

                collected_messages = []

                # Check if module name is openai (so we don't have to import openai package in our library)
                if data.__class__.__module__.startswith("openai"):
                    for chunk in data:
                        content = (
                            chunk.choices[0].delta.content
                            if chunk.choices[0].delta
                            else None
                        )
                        if content:
                            collected_messages.append(content)
                            chunk_info = {
                                **info,
                                "chunk": chunk.model_dump(
                                    exclude_unset=True, exclude_none=True, mode="json"
                                ),
                            }

                            json_message = json.dumps(chunk_info)

                            await websocket.send(message=str(json_message), text=True)
                            heartbeat(json_message)

                    return "".join(collected_messages)
                if isinstance(data, dict):
                    event_info = {
                        **info,
                        "data": data,
                    }

                    json_message = json.dumps(event_info)
                    await websocket.send(message=str(json_message), text=True)
                    heartbeat(json_message)
                    return data
            finally:
                # Ensure the WebSocket connection is closed
                await websocket.close()
    except Exception as e:
        error_message = f"Error with restack stream to websocket: {e}"
        log.exception(error_message)
        raise ApplicationError(error_message) from e



import temporalio.converter
from temporalio.api.common.v1 import Payload, Payloads
from temporalio.converter import DataConverter, PayloadCodec

converter = temporalio.converter

__all__ = ["DataConverter", "Payload", "PayloadCodec", "Payloads", "converter"]
